var ObjectId = require( 'mongoose' ).Schema.ObjectId,
    date_utils = require( 'date-utils' );

// the class
function GenericController ( model ) {
  this.model = model;
}

// export the GenericController class
module.exports = GenericController;

// generic error messages
var error_message = {
  400: 'Bad request',
  401: 'Authorization required',
  403: 'Forbidden',
  404: 'Resource not found',
  409: 'Conflict',
  500: 'Server error'
};

var error_type = {
  400: 'client',
  401: 'client',
  403: 'client',
  404: 'client',
  409: 'client',
  500: 'server'
};

// generic response
GenericController.prototype._respond = function ( res, docs, code, err ) {
  var output = {};

  code = code || 200;

  if ( code > 299 ) {
    if ( ! err ) {
      err = error_message[ code ] || 'Unknown error';
    }
  }

  if ( err ) {
    output.error = {
      code: code,
      type: error_type[ code ] || 'unknown',
      description: err
    };
  }

  if ( docs ) {
    output.data = docs;
  }

  res.send( output, code );
};

// returns the router object that gets passed to
// app.resource()
GenericController.prototype.router = function () {
  // create a closure for each method and expose them each as
  // a property on a plain object
  var self = this;
  return {
    index: function () { self.index.apply( self, arguments ); },
    show: function () { self.show.apply( self, arguments ); },
    new: function () { self.new.apply( self, arguments ); },
    create: function () { self.create.apply( self, arguments ); },
    edit: function () { self.edit.apply( self, arguments ); },
    update: function () { self.update.apply( self, arguments ); },
    destroy: function () { self.destroy.apply( self, arguments ); }
  };
};

// filter the base query based on the request
GenericController.prototype._filter_query = function ( req, query ) {
  var filter = req.param( 'filter' );

  if ( ! filter ) {
    return;
  }

  filter = JSON.parse( filter );

  var field, type, value;
  for ( var i = 0, l = filter.length; i < l; i++ ) {
    field = filter[ i ][ 0 ];
    type = filter[ i ][ 1 ];
    value = filter[ i ][ 2 ];

    switch ( type ) {
      case 'eq':
        query.where( field, value );
      break;

      case 'gt':
        query.where( field ).gt( value );
      break;

      case 'gte':
        query.where( field ).gte( value );
      break;

      case 'in':
        query.where( field ).in( value );
      break;

      case 'lt':
        query.where( field ).lt( value );
      break;

      case 'lte':
        query.where( field ).lte( value );
      break;

      case 'match':
        query.where( field, new RegExp( value ) );
      break;

      case 'on':
        var y = value[ 0 ],
            m = ( value[ 1 ] - 1 ) || 0,
            d = value[ 2 ] || 1,
            add = {};

        if ( value.length === 1 ) {
          add.years = 1;
        } else if ( value.length === 2 ) {
          add.months = 1;
        } else {
          add.days = 1;
        }

        var before = new Date( y, m, d ),
            after = new Date( y, m, d ).add( add );

        query.where( field ).gte( before ).lte( after );
      break;
    }
  }
};

// create the base query used to list instances of the model
GenericController.prototype._index_query = function ( req ) {
  var query = this.model.find();

  if ( req.param( 'sort' ) ) {
    var sort_dir = ( req.param( 'dir' ) === 'asc' ) ? 'ascending' : 'descending';
    query.sort( req.param( 'sort' ), sort_dir );
  }

  if ( req.param( 'start' ) ) {
    query.skip( Number( req.param( 'start' ) ) );
  }

  if ( req.param( 'limit' ) ) {
    query.limit( Number( req.param( 'limit' ) ) );
  }

  this._filter_query( req, query );

  return query;
};

// attempt to fetch the items specified in the request
// if the attempt fails, respond with an error
GenericController.prototype._index_fetch = function ( req, res, query_mod, cb ) {
  var query = this._index_query( req ),
      respond = this._respond;

  if ( query_mod ) {
    query_mod( query );
  }

  query.run( function ( err, docs ) {
    if ( err ) {
      respond( res, null, 500 );
    } else {
      cb( docs );
    }
  });
};

// list instances of the model
GenericController.prototype.index = function ( req, res ) {
  var self = this;
  this._index_fetch( req, res, null, function ( docs ) {
    self._respond( res, docs );
  });
};

// generate a query for a particular item
GenericController.prototype._item_query = function ( req ) {
  return this.model.findById( this._item_id( req ) );
};

// attempt to fetch the item specified in the request
// if the attempt fails, respond with an error
GenericController.prototype._item_fetch = function ( req, res, query_mod, cb ) {
  var query = this._item_query( req ),
      respond = this._respond;

  if ( query_mod ) {
    query_mod( query );
  }

  query.run( function ( err, doc ) {
    if ( err && err.message === 'Invalid ObjectId' ) {
      respond( res, null, 404 );
    } else if ( err ) {
      respond( res, null, 500 );
    } else {
      cb( doc );
    }
  });
};

// show a particular instance
GenericController.prototype.show = function ( req, res ) {
  var self = this;
  this._item_fetch( req, res, null, function ( item ) {
    self._respond( res, item );
  });
};

GenericController.prototype._validation_response = function ( res, err ) {
  var message = err.message,
      self = this;

  function unknown() {
    self._respond( res, null, 500, 'Unknown error' );
  }

  if ( message ) {
    if ( message === 'created lock' ) {
      self._respond( res, null, 403, 'Created date may not be set manually' );
    } else if ( message === 'updated lock' ) {
      self._respond( res, null, 403, 'Updated date may not be set manually' );
    } else {
      unknown();
    }
  } else {
    unknown();
  }
};

// attempt to create an item from the request
// if the attempt fails, respond with an error
GenericController.prototype._item_create = function ( req, res, cb ) {
  var self = this,
      item = new this.model( req.body );

  item.save( function ( err, item ) {
    if ( err ) {
      self._validation_response( res, err );
    } else {
      cb( item );
    }
  });

};

// create an item from the request
GenericController.prototype.create = function ( req, res ) {
  var self = this;
  this._item_create( req, res, function ( item ) {
    // TODO set Location header
    // res.header( 'Location', blah );
    self._respond( res, item, 201 );
  });
};

// attempt to update the item from the request
// if the attempt fails, respond with an error
GenericController.prototype._item_update = function ( req, res, cb ) {
  var self = this,
      item_id = this._item_id( req );

  this._item_fetch( req, res, null, function ( item ) {
    for ( var key in req.body ) {
      item[ key ] = req.body[ key ];
    }

    item.save( function ( err ) {
      if ( err ) {
        self._validation_response( res, err );
      } else {
        cb( item );
      }
    });
  });
};

// update the given item
GenericController.prototype.update = function ( req, res ) {
  var self = this;
  this._item_update( req, res, function ( item ) {
    self._respond( res, item );
  });
};

GenericController.prototype.new = function ( req, res ) {
  this._respond( res, null, 400 );
};

GenericController.prototype.edit = function ( req, res ) {
  this._respond( res, null, 400 );
};

GenericController.prototype.destroy = function ( req, res ) {
  var self = this;

  this._item_fetch( req, res, null, function ( item ) {
    item.remove( function ( err ) {
      if ( err ) {
        self._respond( res, null, 500 );
      } else {
        self._respond( res, null, 200 );
      }
    });
  });
};

