var vows = require( 'vows' ),
    assert = require( 'assert' ),
    querystring = require( 'querystring' ),

    lib = require( './lib' ),
    request = lib.request,

    Author = require( '../models/Author' );

function setup ( cb ) {
  lib.setup( function ( conf ) {

  var author1 = new Author({
    name: 'Test Author B',
    url: 'test-author-b'
  });

  var author2 = new Author({
    name: 'Test Author A',
    url: 'test-author-a'
  });

  author1.save( function () {
    author2.save( function () {
      conf.author1 = author1;
      conf.author2 = author2;
      cb( conf );
    });
  });

  });
}

setup( function ( conf ) {

var userpass = [ conf.user, conf.password ];

function verify( data ) {
  var i = 0, found = 0;
  for ( ; i < data.length; i ++ ) {
    if ( data[ i ].url.substring( 0, 12 ) === 'test-author-' ) {
      found++;
    }
  }
  return found;
}

vows.describe( 'Author Tests' ).addBatch({

  'WHEN I list authors without authenticating': {
    topic: function () {
      request( null, { uri: '/authors' }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I list authors': {
    topic: function () {
      request( userpass, { uri: '/authors' }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      var found = verify( body.data );

      assert.equal( found, 2 );
      assert.equal( body.data.length, 2 );
    }
  },

  'WHEN I list authors sorted': {
    topic: function () {
      request( userpass, {
        uri: '/authors?sort=url&dir=desc'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      var found = verify( body.data );

      assert.equal( found, 2 );
      assert.equal( body.data.length, 2 );
    },
    'THEN the order should be correct': function ( err, res, body ) {
      body = JSON.parse( body );

      for ( var i = 0; i < body.data.length; i++ ) {
        if ( i === 0 ) {
          continue;
        } else if ( body.data[ i - 1 ].url < body.data[ i ].url ) {
          assert.fail();
        }
      }
    }
  },

  'WHEN I list authors sorted by created date': {
    topic: function () {
      request( userpass, {
        uri: '/authors?sort=created&dir=asc'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN the data should be correct': function ( err, res, body ) {
      body = JSON.parse( body );

      var found = verify( body.data );

      assert.equal( found, 2 );
      assert.equal( body.data.length, 2 );
    },
    'THEN the data should be sorted correctly': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data[ 0 ]._id, conf.author1._id );
      assert.equal( body.data[ 1 ]._id, conf.author2._id );
    }
  },

  'WHEN I query sorted and starting at item 1': {
    topic: function () {
      request( userpass, {
        uri: '/authors?sort=name&dir=desc&start=1'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data.length, 1 );
      assert.equal( body.data[ 0 ].name, 'Test Author A' );
    }
  },

  'WHEN I query with a limit': {
    topic: function () {
      request( userpass, {
        uri: '/authors?limit=1'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get only one item': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data.length, 1 );
    }
  },

  'WHEN I query with eq': {
    topic: function () {
      var query = querystring.stringify({
        filter: JSON.stringify([
          [ '_id', 'eq', conf.author2._id ]
        ])
      });
      request( userpass, {
        uri: '/authors?' + query
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data[ 0 ]._id, conf.author2._id );
    }
  },

  'WHEN I query with gt': {
    topic: function () {
      var query = querystring.stringify({
        filter: JSON.stringify([
          [ 'created', 'gt', conf.author1.created ]
        ])
      });
      request( userpass, {
        uri: '/authors?' + query
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data[ 0 ]._id, conf.author2._id );
    }
  },

  'WHEN I query with gte': {
    topic: function () {
      var query = querystring.stringify({
        filter: JSON.stringify([
          [ 'created', 'gte', conf.author1.created ]
        ]),
        sort: 'created',
        dir: 'asc'
      });
      request( userpass, {
        uri: '/authors?' + query
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data[ 0 ]._id, conf.author1._id );
      assert.equal( body.data[ 1 ]._id, conf.author2._id );
    }
  },

  'WHEN I query with in': {
    topic: function () {
      var query = querystring.stringify({
        filter: JSON.stringify([
          [ 'url', 'in', [ 'beanz', 'test-author-b' ] ]
        ])
      });
      request( userpass, {
        uri: '/authors?' + query
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data[ 0 ]._id, conf.author1._id );
    }
  },

  'WHEN I query with lt': {
    topic: function () {
      var query = querystring.stringify({
        filter: JSON.stringify([
          [ 'created', 'lt', conf.author2.created ]
        ])
      });
      request( userpass, {
        uri: '/authors?' + query
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data[ 0 ]._id, conf.author1._id );
    }
  },

  'WHEN I query with lte': {
    topic: function () {
      var query = querystring.stringify({
        filter: JSON.stringify([
          [ 'created', 'lte', conf.author2.created ]
        ]),
        sort: 'created',
        dir: 'asc'
      });
      request( userpass, {
        uri: '/authors?' + query
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data[ 0 ]._id, conf.author1._id );
      assert.equal( body.data[ 1 ]._id, conf.author2._id );
    }
  },


  'WHEN I query with match': {
    topic: function () {
      var query = querystring.stringify({
        filter: JSON.stringify([
          [ 'name', 'match', 'Author A' ]
        ])
      });
      request( userpass, {
        uri: '/authors?' + query
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data[ 0 ]._id, conf.author2._id );
    }
  },

  'WHEN I request one author without authenticating': {
    topic: function () {
      request( null, { uri: '/authors/' + conf.author1._id }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I request one author that doesn\'t exist': {
    topic: function () {
      request( userpass, {
        uri: '/authors/abcd1234'
      }, this.callback );
    },
    'THEN I should get a 404': function ( err, res, body ) {
      assert.equal( res.statusCode, 404 );
    }
  },

  'WHEN I request one author': {
    topic: function () {
      request( userpass, {
        uri: '/authors/' + conf.author1._id
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct author': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data._id, conf.author1._id );
      assert.equal( body.data.name, conf.author1.name );
      assert.isUndefined( body.data.password );
      assert.notEqual( String( new Date( body.data.created ) ), 'Invalid Date' );
      assert.notEqual( String( new Date( body.data.updated ) ), 'Invalid Date' );
    }
  },

  'WHEN I create an author without authenticating': {
    topic: function () {
      request( null, {
        uri: '/authors',
        method: 'POST',
        body: JSON.stringify({
          name: 'bean',
          url: 'bean'
        })
      }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I create an author with a conflicting url': {
    topic: function () {
      request( userpass, {
        uri: '/authors',
        method: 'POST',
        body: JSON.stringify({
          name: 'bean',
          url: 'test-author-a'
        })
      }, this.callback );
    },
    'THEN I should get a 409': function ( err, res, body ) {
      assert.equal( res.statusCode, 409 );
    }
  },

  'WHEN I create an author with no name': {
    topic: function () {
      request( userpass, {
        uri: '/authors',
        method: 'POST',
        body: JSON.stringify({
          url: 'bean'
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I create an author with no url': {
    topic: function () {
      request( userpass, {
        uri: '/authors',
        method: 'POST',
        body: JSON.stringify({
          name: 'bean'
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I create an author with a created date': {
    topic: function () {
      var now = Date.now();
      request( userpass, {
        uri: '/authors',
        method: 'POST',
        body: JSON.stringify({
          name: 'bean',
          url: 'bean',
          created: now
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I create an author with an updated date': {
    topic: function () {
      var now = Date.now();
      request( userpass, {
        uri: '/authors',
        method: 'POST',
        body: JSON.stringify({
          name: 'bean',
          url: 'bean',
          updated: now
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I create an author with an invalid name': {
    topic: function () {
      request( userpass, {
        uri: '/authors',
        method: 'POST',
        body: JSON.stringify({
          name: '',
          url: 'george'
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I create an author with an invalid url': {
    topic: function () {
      request( userpass, {
        uri: '/authors',
        method: 'POST',
        body: JSON.stringify({
          name: 'george'
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I create an author': {
    topic: function () {
      request( userpass, {
        uri: '/authors',
        method: 'POST',
        body: JSON.stringify({
          name: 'bean',
          url: 'bean'
        })
      }, this.callback );
    },
    'THEN I should get a 201': function ( err, res, body ) {
      assert.equal( res.statusCode, 201 );
    },
    'THEN I should get the author back': function ( err, res, body ) {
      body = JSON.parse( body );

      created_id = body.data._id;

      assert.isNotNull( body.data._id );
      assert.equal( body.data.name, 'bean' );
      assert.equal( body.data.url, 'bean' );
      assert.notEqual( String( new Date( body.data.created ) ), 'Invalid Date' );
      assert.notEqual( String( new Date( body.data.updated ) ), 'Invalid Date' );
    }
  }

}).addBatch({

  'WHEN I verify that the author was created': {
    topic: function () {
      Author.findById( created_id )
        .run( this.callback );
    },
    'THEN I should get the correct data': function ( err, author ) {
      assert.equal( author._id, created_id );
      assert.equal( author.name, 'bean' );
      assert.equal( author.url, 'bean' );
      assert.isTrue( author.created instanceof Date );
      assert.isTrue( author.updated instanceof Date );
    }
  }

}).addBatch({

  'WHEN I update an author without authenticating': {
    topic: function () {
      request( null, {
        uri: '/authors/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          name: 'George'
        })
      }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I update an author that doesn\'t exist': {
    topic: function () {
      request( userpass, {
        uri: '/authors/1234',
        method: 'PUT',
        body: JSON.stringify({
          name: 'George'
        })
      }, this.callback );
    },
    'THEN I should get a 404': function ( err, res, body ) {
      assert.equal( res.statusCode, 404 );
    }
  },

  'WHEN I update an author with a created date': {
    topic: function () {
      var now = Date.now();
      request( userpass, {
        uri: '/authors/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          created: now
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I update an author with an updated date': {
    topic: function () {
      var now = Date.now();
      request( userpass, {
        uri: '/authors/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          updated: now
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I update an author with an invalid name': {
    topic: function () {
      request( userpass, {
        uri: '/authors/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          name: ''
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I update an author with an invalid url': {
    topic: function () {
      request( userpass, {
        uri: '/authors/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          url: ''
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I update an author with a conflicting url': {
    topic: function () {
      request( userpass, {
        uri: '/authors/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          url: 'test-author-a'
        })
      }, this.callback );
    },
    'THEN I should get a 409': function ( err, res, body ) {
      assert.equal( res.statusCode, 409 );
    }
  },

  'WHEN I update an author': {
    topic: function () {
      request( userpass, {
        uri: '/authors/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          name: 'George',
          url: 'george'
        })
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data back': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data._id, created_id );
      assert.equal( body.data.name, 'George' );
      assert.equal( body.data.url, 'george' );
    }
  }

}).addBatch({

  // TODO verify that updated date was really updated

  'WHEN I verify that the author was updated': {
    topic: function () {
      Author.findById( created_id )
        .run( this.callback );
    },
    'THEN I should get the correct data': function ( err, author ) {
      assert.equal( author._id, created_id );
      assert.equal( author.name, 'George' );
      assert.equal( author.url, 'george' );
      assert.isTrue( author.created instanceof Date );
      assert.isTrue( author.updated instanceof Date );
    }
  }

}).addBatch({

  'WHEN I delete an author without authenticating': {
    topic: function () {
      request( null, {
        uri: '/authors/' + created_id,
        method: 'DELETE',
      }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I delete an author that doesn\'t exist': {
    topic: function () {
      request( userpass, {
        uri: '/authors/abcd1234',
        method: 'DELETE'
      }, this.callback );
    },
    'THEN I should get a 404': function ( err, res, body ) {
      assert.equal( res.statusCode, 404 );
    }
  },

  'WHEN I delete an author': {
    topic: function () {
      request( userpass, {
        uri: '/authors/' + created_id,
        method: 'DELETE'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    }
  }

}).addBatch({

  'WHEN I verify that the author was deleted': {
    topic: function () {
      Author.findById( created_id, this.callback );
    },
    'THEN it should be gone': function ( err, doc ) {
      assert.isNull( doc );
    }
  },

  teardown: conf.teardown

}).export( module );

});
