var vows = require( 'vows' ),
    assert = require( 'assert' ),
    lib = require( './lib' ),
    request = lib.request,
    Blog = require( '../models/Blog' );

function setup ( cb ) {
  lib.setup( function ( conf ) {

  var blog1 = new Blog({
    title: 'Test Blog A'
  });

  var blog2 = new Blog({
    title: 'Test Blog B'
  });

  blog1.save( function () {
    blog2.save( function () {
      conf.blog1 = blog1;
      conf.blog2 = blog2;
      cb( conf );
    });
  });

  });
}

setup( function ( conf ) {

var userpass = [ conf.user, conf.password ];

function verify( data ) {
  var i = 0, found = 0;
  for ( ; i < data.length; i ++ ) {
    if ( data[ i ].title.substring( 0, 10 ) === 'Test Blog ' ) {
      found++;
    }
  }
  return found;
}

vows.describe( 'Blog Tests' ).addBatch({

  'WHEN I list blogs without authenticating': {
    topic: function () {
      request( null, { uri: '/blogs' }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I list blogs': {
    topic: function () {
      request( userpass, { uri: '/blogs' }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      var found = verify( body.data );

      assert.equal( found, 2 );
      assert.equal( body.data.length, 2 );
    }
  },

  'WHEN I list blogs sorted': {
    topic: function () {
      request( userpass, {
        uri: '/blogs?sort=title&dir=desc'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      var found = verify( body.data );

      assert.equal( found, 2 );
      assert.equal( body.data.length, 2 );
    },
    'THEN the order should be correct': function ( err, res, body ) {
      body = JSON.parse( body );

      for ( var i = 0; i < body.data.length; i++ ) {
        if ( i === 0 ) {
          continue;
        } else if ( body.data[ i - 1 ].url < body.data[ i ].url ) {
          assert.fail();
        }
      }
    }
  },

  'WHEN I list blogs sorted by created date': {
    topic: function () {
      request( userpass, {
        uri: '/blogs?sort=created&dir=asc'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN the data should be correct': function ( err, res, body ) {
      body = JSON.parse( body );

      var found = verify( body.data );

      assert.equal( found, 2 );
      assert.equal( body.data.length, 2 );
    },
    'THEN the data should be sorted correctly': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data[ 0 ]._id, conf.blog1._id );
      assert.equal( body.data[ 1 ]._id, conf.blog2._id );
    }
  },


  'WHEN I query sorted and starting at item 1': {
    topic: function () {
      request( userpass, {
        uri: '/blogs?sort=title&dir=desc&start=1'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data.length, 1 );
      assert.equal( body.data[ 0 ].title, 'Test Blog A' );
    }
  },

  'WHEN I query with a limit': {
    topic: function () {
      request( userpass, {
        uri: '/blogs?limit=1'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get only one item': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data.length, 1 );
    }
  },

  'WHEN I request one blog without authenticating': {
    topic: function () {
      request( null, { uri: '/blogs/' + conf.blog1._id }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I request one blog that doesn\'t exist': {
    topic: function () {
      request( userpass, {
        uri: '/blogs/abcd1234'
      }, this.callback );
    },
    'THEN I should get a 404': function ( err, res, body ) {
      assert.equal( res.statusCode, 404 );
    }
  },

  'WHEN I request one blog': {
    topic: function () {
      request( userpass, {
        uri: '/blogs/' + conf.blog1._id
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct blog': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data._id, conf.blog1._id );
      assert.equal( body.data.title, conf.blog1.title );
      assert.notEqual( String( new Date( body.data.created ) ), 'Invalid Date' );
      assert.notEqual( String( new Date( body.data.updated ) ), 'Invalid Date' );
    }
  },

  'WHEN I create a blog without authenticating': {
    topic: function () {
      request( null, {
        uri: '/blogs',
        method: 'POST',
        body: JSON.stringify({
          name: 'title'
        })
      }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I create a blog with no title': {
    topic: function () {
      request( userpass, {
        uri: '/blogs',
        method: 'POST',
        body: JSON.stringify({
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I create a blog with a created date': {
    topic: function () {
      var now = Date.now();
      request( userpass, {
        uri: '/blogs',
        method: 'POST',
        body: JSON.stringify({
          title: 'bean',
          created: now
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I create a blog with an updated date': {
    topic: function () {
      var now = Date.now();
      request( userpass, {
        uri: '/blogs',
        method: 'POST',
        body: JSON.stringify({
          title: 'bean',
          updated: now
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I create a blog': {
    topic: function () {
      request( userpass, {
        uri: '/blogs',
        method: 'POST',
        body: JSON.stringify({
          title: 'title'
        })
      }, this.callback );
    },
    'THEN I should get a 201': function ( err, res, body ) {
      assert.equal( res.statusCode, 201 );
    },
    'THEN I should get the blog back': function ( err, res, body ) {
      body = JSON.parse( body );

      created_id = body.data._id;

      assert.isNotNull( body.data._id );
      assert.equal( body.data.title, 'title' );
      assert.notEqual( String( new Date( body.data.created ) ), 'Invalid Date' );
      assert.notEqual( String( new Date( body.data.updated ) ), 'Invalid Date' );
    }
  }

}).addBatch({

  'WHEN I verify that the blog was created': {
    topic: function () {
      Blog.findById( created_id )
        .run( this.callback );
    },
    'THEN I should get the correct data': function ( err, blog ) {
      assert.equal( blog._id, created_id );
      assert.equal( blog.title, 'title' );
      assert.isTrue( blog.created instanceof Date );
      assert.isTrue( blog.updated instanceof Date );
    }
  }

}).addBatch({

  'WHEN I update a blog without authenticating': {
    topic: function () {
      request( null, {
        uri: '/blogs/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          title: 'George'
        })
      }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I update a blog that doesn\'t exist': {
    topic: function () {
      request( userpass, {
        uri: '/blogs/abcd1234',
        method: 'PUT',
        body: JSON.stringify({
          title: 'George'
        })
      }, this.callback );
    }
  },

  'WHEN I update a blog with an invalid title': {
    topic: function () {
      request( userpass, {
        uri: '/blogs/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          title: ''
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I update a blog with a created date': {
    topic: function () {
      var now = Date.now();
      request( userpass, {
        uri: '/blogs/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          created: now
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I update a blog with an updated date': {
    topic: function () {
      var now = Date.now();
      request( userpass, {
        uri: '/blogs/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          updated: now
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I update a blog': {
    topic: function () {
      request( userpass, {
        uri: '/blogs/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          title: 'George'
        })
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data back': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data._id, created_id );
      assert.equal( body.data.title, 'George' );
    }
  }

}).addBatch({

  // TODO verify that updated date was really updated

  'WHEN I verify that the blog was updated': {
    topic: function () {
      Blog.findById( created_id )
        .run( this.callback );
    },
    'THEN I should get the correct data': function ( err, blog ) {
      assert.equal( blog._id, created_id );
      assert.equal( blog.title, 'George' );
    }
  }

}).addBatch({

  'WHEN I delete a blog without authenticating': {
    topic: function () {
      request( null, {
        uri: '/blogs/' + created_id,
        method: 'DELETE',
      }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I delete a blog that doesn\'t exist': {
    topic: function () {
      request( userpass, {
        uri: '/blogs/abcd1234',
        method: 'DELETE'
      }, this.callback );
    },
    'THEN I should get a 404': function ( err, res, body ) {
      assert.equal( res.statusCode, 404 );
    }
  },

  'WHEN I delete a blog': {
    topic: function () {
      request( userpass, {
        uri: '/blogs/' + created_id,
        method: 'DELETE'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    }
  }

}).addBatch({

  'WHEN I verify that the blog was deleted': {
    topic: function () {
      Blog.findById( created_id, this.callback );
    },
    'THEN it should be gone': function ( err, doc ) {
      assert.isNull( doc );
    }
  },

  teardown: conf.teardown

}).export( module );

});
