var vows = require( 'vows' ),
    assert = require( 'assert' ),
    lib = require( './lib' ),
    request = lib.request,
    async = require( 'async' ),

    Author = require( '../models/Author' ),
    Blog = require( '../models/Blog' ),
    Tag = require( '../models/Tag' ),
    Post = require( '../models/Post' );

function setup ( cb ) {
  lib.setup( function ( conf ) {

  var author1 = new Author({
    name: 'testauthor1',
    url: 'test-author-1'
  });

  var author2 = new Author({
    name: 'testauthor2',
    url: 'test-author-2'
  });

  var blog1 = new Blog({
    title: 'blog1'
  });

  var blog2 = new Blog({
    title: 'blog2'
  });

  async.parallel([
    function ( cb ) { blog1.save( cb ); },
    function ( cb ) { blog2.save( cb ); },
    function ( cb ) { author1.save( cb ); },
    function ( cb ) { author2.save( cb ); }
  ], function () {

    var tag1 = new Tag({
      url: 'tag1',
      title: 'Tag 1',
      blog: blog1._id
    });

    var tag2 = new Tag({
      url: 'tag2',
      title: 'Tag 2',
      blog: blog1._id
    });

    async.parallel([
      function ( cb ) { tag1.save( cb ); },
      function ( cb ) { tag2.save( cb ); }
    ], function () {

      var post1 = new Post({
        url: 'testpost1',
        title: 'Test Title',
        body: 'Test body',
        authors: [ author1, author2 ],
        tags: [ tag1, tag2 ]
      });

      var post2 = new Post({
        url: 'testpost2',
        title: 'Test Title',
        body: 'Test body',
        authors: [ author1 ]
      });

      post1.save( function () {
      post2.save( function () {

        conf.blog1 = blog1;
        conf.blog2 = blog2;
        conf.author1 = author1;
        conf.author2 = author2;
        conf.tag1 = tag1;
        conf.tag2 = tag2;
        conf.post1 = post1;
        conf.post2 = post2;

        cb( conf );
      }); });
    });
  });

  });
}

setup( function ( conf ) {

var userpass = [ conf.user, conf.password ],
  created_id;

function verify( data ) {
  var i = 0, found = 0;
  for ( ; i < data.length; i ++ ) {
    if ( data[ i ].url.substring( 0, 8 ) === 'testpost' ) {
      found++;
    }
  }
  return found;
}

vows.describe( 'Post Tests' ).addBatch({

  'WHEN I list posts without authenticating': {
    topic: function () {
      request( null, { uri: '/posts' }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I list posts': {
    topic: function () {
      request( userpass, { uri: '/posts' }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      var found = verify( body.data );

      assert.equal( found, 2 );
      assert.equal( body.data.length, 2 );
    }
  },

  'WHEN I list posts sorted': {
    topic: function () {
      request( userpass, {
        uri: '/posts?sort=url&dir=desc'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      var found = verify( body.data );

      assert.equal( found, 2 );
      assert.equal( body.data.length, 2 );
    },
    'THEN the order should be correct': function ( err, res, body ) {
      body = JSON.parse( body );

      for ( var i = 0; i < body.data.length; i++ ) {
        if ( i === 0 ) {
          continue;
        } else if ( body.data[ i - 1 ].url < body.data[ i ].url ) {
          assert.fail();
        }
      }
    }
  },

  'WHEN I list posts sorted by created date': {
    topic: function () {
      request( userpass, {
        uri: '/posts?sort=created&dir=asc'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN the data should be correct': function ( err, res, body ) {
      body = JSON.parse( body );

      var found = verify( body.data );

      assert.equal( found, 2 );
      assert.equal( body.data.length, 2 );
    },
    'THEN the data should be sorted correctly': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( String( body.data[ 0 ]._id ), String( conf.post1._id ) );
      assert.equal( String( body.data[ 1 ]._id ), String( conf.post2._id ) );
    }
  },

  'WHEN I query sorted and starting at item 1': {
    topic: function () {
      request( userpass, {
        uri: '/posts?sort=url&dir=asc&start=1'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data.length, 1 );
      assert.equal( body.data[ 0 ].url, 'testpost2' );
    }
  },

  'WHEN I query with a limit': {
    topic: function () {
      request( userpass, {
        uri: '/posts?limit=1'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get only one item': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data.length, 1 );
    }
  },

  'WHEN I request one post without authenticating': {
    topic: function () {
      request( null, { uri: '/posts/' + conf.post1._id }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I request one post that doesn\'t exist': {
    topic: function () {
      request( userpass, {
        uri: '/posts/abcd1234'
      }, this.callback );
    },
    'THEN I should get a 404': function ( err, res, body ) {
      assert.equal( res.statusCode, 404 );
    }
  },

  'WHEN I request one post': {
    topic: function () {
      request( userpass, {
        uri: '/posts/' + conf.post1._id
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct post': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( String( body.data._id ), String( conf.post1._id ) );
      assert.equal( body.data.title, conf.post1.title );
      assert.equal( body.data.body, conf.post1.body );
      assert.equal( body.data.url, conf.post1.url );

      assert.equal( body.data.authors.length, 2 );
      assert.equal( String( body.data.authors[ 0 ] ), String( conf.author1._id ) );
      assert.equal( String( body.data.authors[ 1 ] ), String( conf.author2._id ) );

      assert.equal( body.data.tags.length, 2 );
      assert.equal( String( body.data.tags[ 0 ] ), String( conf.tag1._id ) );
      assert.equal( String( body.data.tags[ 1 ] ), String( conf.tag2._id ) );

      assert.notEqual( String( new Date( body.data.created ) ), 'Invalid Date' );
      assert.notEqual( String( new Date( body.data.updated ) ), 'Invalid Date' );
    }
  },

  'WHEN I create a post without authenticating': {
    topic: function () {
      request( null, {
        uri: '/posts',
        method: 'POST',
        body: JSON.stringify({
          title: 'Blah Title',
          body: 'Blah body',
          url: 'blahurl',
          pubdate: Date.now(),
          authors: [ conf.author1._id ],
          tags: [ conf.tag1._id ]
        })
      }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I create a post with no title': {
    topic: function () {
      request( userpass, {
        uri: '/posts',
        method: 'POST',
        body: JSON.stringify({
          body: 'Blah body',
          url: 'blahurl',
          pubdate: Date.now(),
          authors: [ conf.author1._id ],
          tags: [ conf.tag1._id ]
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I create a post with no body': {
    topic: function () {
      request( userpass, {
        uri: '/posts',
        method: 'POST',
        body: JSON.stringify({
          title: 'Blah title',
          url: 'blahurl',
          pubdate: Date.now(),
          authors: [ conf.author1._id ],
          tags: [ conf.tag1._id ]
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I create a post with no url': {
    topic: function () {
      request( userpass, {
        uri: '/posts',
        method: 'POST',
        body: JSON.stringify({
          title: 'Blah title',
          body: 'Blah body',
          pubdate: Date.now(),
          authors: [ conf.author1._id ],
          tags: [ conf.tag1._id ]
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  // TODO test with invalid authors and tags

  'WHEN I create a post': {
    topic: function () {
      request( userpass, {
        uri: '/posts',
        method: 'POST',
        body: JSON.stringify({
          title: 'Blah title',
          body: 'Blah body',
          url: 'blahurl',
          pubdate: '2011-01-01',
          authors: [ conf.author1._id ],
          tags: [ conf.tag1._id ]
        })
      }, this.callback );
    },
    'THEN I should get a 201': function ( err, res, body ) {
      assert.equal( res.statusCode, 201 );
    },
    'THEN I should get the post back': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.isNotNull( body.data._id );

      created_id = body.data._id;

      assert.equal( body.data.title, 'Blah title' );
      assert.equal( body.data.body, 'Blah body' );
      assert.equal( body.data.url, 'blahurl' );

      assert.equal( body.data.authors.length, 1 );
      assert.equal( String( body.data.authors[ 0 ] ), String( conf.author1._id ) );

      assert.equal( body.data.tags.length, 1 );
      assert.equal( String( body.data.tags[ 0 ] ), String( conf.tag1._id ) );

      assert.notEqual( String( new Date( body.data.created ) ), 'Invalid Date' );
      assert.notEqual( String( new Date( body.data.updated ) ), 'Invalid Date' );
      assert.notEqual( String( new Date( body.data.pubdate ) ), 'Invalid Date' );
    }
  },

}).addBatch({

  'WHEN I verify that the post was created': {
    topic: function () {
      Post.findById( created_id )
        .populate( 'authors' )
        .populate( 'tags' )
        .run( this.callback );
    },
    'THEN I should get the correct data': function ( err, post ) {
      assert.equal( post._id, created_id );
      assert.equal( post.title, 'Blah title' );
      assert.equal( post.body, 'Blah body' );
      assert.equal( post.url, 'blahurl' );

      assert.equal( post.pubdate.getTime(), new Date( '2011-01-01' ).getTime() );
      assert.equal( String( post.tags[ 0 ]._id ), String( conf.tag1._id ) );
      assert.equal( String( post.authors[ 0 ]._id ), String( conf.author1._id ) );
    }
  },

}).addBatch({

  'WHEN I update a post without authenticating': {
    topic: function () {
      request( null, {
        uri: '/posts/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          url: 'george'
        })
      }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I update a post that doesn\'t exist': {
    topic: function () {
      request( userpass, {
        uri: '/posts/1234',
        method: 'PUT',
        body: JSON.stringify({
          url: 'george'
        })
      }, this.callback );
    },
    'THEN I should get a 404': function ( err, res, body ) {
      assert.equal( res.statusCode, 404 );
    }
  },

  'WHEN I update a post with an invalid url': {
    topic: function () {
      request( userpass, {
        uri: '/posts/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          url: ''
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I update a post with an invalid title': {
    topic: function () {
      request( userpass, {
        uri: '/posts/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          title: ''
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I update a post with an invalid body': {
    topic: function () {
      request( userpass, {
        uri: '/posts/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          body: ''
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I update a post with a created date': {
    topic: function () {
      var now = Date.now();
      request( userpass, {
        uri: '/posts/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          created: now
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I update a post with an updated date': {
    topic: function () {
      var now = Date.now();
      request( userpass, {
        uri: '/posts/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          updated: now
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I update a post': {
    topic: function () {
      request( userpass, {
        uri: '/posts/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          url: 'george'
        })
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data back': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data._id, created_id );
      assert.equal( body.data.url, 'george' );
    }
  }

}).addBatch({

  // TODO verify that updated date was really updated

  'WHEN I verify that the post was updated': {
    topic: function () {
      Post.findById( created_id )
        .run( this.callback );
    },
    'THEN I should get the correct data': function ( err, post ) {
      assert.equal( post._id, created_id );
      assert.equal( post.url, 'george' );
    }
  },

}).addBatch({

  'WHEN I delete a post without authenticating': {
    topic: function () {
      request( null, {
        uri: '/posts/' + created_id,
        method: 'DELETE',
      }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I delete a post that doesn\'t exist': {
    topic: function () {
      request( userpass, {
        uri: '/posts/abcd1234',
        method: 'DELETE'
      }, this.callback );
    },
    'THEN I should get a 404': function ( err, res, body ) {
      assert.equal( res.statusCode, 404 );
    }
  },

  'WHEN I delete a post': {
    topic: function () {
      request( userpass, {
        uri: '/posts/' + created_id,
        method: 'DELETE'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    }
  }

}).addBatch({

  'WHEN I verify that the post was deleted': {
    topic: function () {
      Post.findById( created_id, this.callback );
    },
    'THEN it should be gone': function ( err, doc ) {
      assert.isNull( doc );
    }
  },

  teardown: conf.teardown

}).export( module );

});

