var vows = require( 'vows' ),
    assert = require( 'assert' ),
    lib = require( './lib' ),
    request = lib.request,
    async = require( 'async' ),

    Blog = require( '../models/Blog' ),
    Tag = require( '../models/Tag' );

function setup ( cb ) {
  lib.setup( function ( conf ) {

  var blog1 = new Blog({
    title: 'Test Blog'
  });

  blog1.save( function () {

    var tag1 = new Tag({
      blog: blog1._id,
      url: 'tag1',
      title: 'Happy Tag'
    });

    var tag2 = new Tag({
      blog: blog1._id,
      url: 'tag2',
      title: 'Sad Tag'
    });

    tag1.save( function () {
      tag2.save( function () {
        conf.blog1 = blog1;
        conf.tag1 = tag1;
        conf.tag2 = tag2;

        cb( conf );
      });
    });
  });

  });
}

setup( function ( conf ) {

var userpass = [ conf.user, conf.password ],
  created_id;

function verify( data ) {
  var i = 0, found = 0;
  for ( ; i < data.length; i ++ ) {
    if ( ( data[ i ].url === 'tag1' ) || 
       ( data[ i ].url === 'tag2' ) ) {
      found++;
    }
    if ( String( data[ i ].blog ) !== String( conf.blog1._id ) ) {
      assert.fail();
    }
  }
  return found;
}

vows.describe( 'Tag Tests' ).addBatch({

  'WHEN I list tags without authenticating': {
    topic: function () {
      request( null, { uri: '/tags' }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I list tags': {
    topic: function () {
      request( userpass, { uri: '/tags' }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      var found = verify( body.data );

      assert.equal( found, 2 );
      assert.equal( body.data.length, 2 );
    }
  },

  'WHEN I list tags sorted': {
    topic: function () {
      request( userpass, {
        uri: '/tags?sort=url&dir=desc'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      var found = verify( body.data );

      assert.equal( found, 2 );
      assert.equal( body.data.length, 2 );
    },
    'THEN the order should be correct': function ( err, res, body ) {
      body = JSON.parse( body );

      for ( var i = 0; i < body.data.length; i++ ) {
        if ( i === 0 ) {
          continue;
        } else if ( body.data[ i - 1 ].url < body.data[ i ].url ) {
          assert.fail();
        }
      }
    }
  },

  'WHEN I list tags sorted by created date': {
    topic: function () {
      request( userpass, {
        uri: '/tags?sort=created&dir=asc'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN the data should be correct': function ( err, res, body ) {
      body = JSON.parse( body );

      var found = verify( body.data );

      assert.equal( found, 2 );
      assert.equal( body.data.length, 2 );
    },
    'THEN the data should be sorted correctly': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( String( body.data[ 0 ]._id ), String( conf.tag1._id ) );
      assert.equal( String( body.data[ 1 ]._id ), String( conf.tag2._id ) );
    }
  },


  'WHEN I query sorted and starting at item 1': {
    topic: function () {
      request( userpass, {
        uri: '/tags?sort=url&dir=asc&start=1'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data.length, 1 );
      assert.equal( body.data[ 0 ].url, 'tag2' );
    }
  },

  'WHEN I query with a limit': {
    topic: function () {
      request( userpass, {
        uri: '/tags?limit=1'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get only one item': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data.length, 1 );
    }
  },

  'WHEN I request one tag without authenticating': {
    topic: function () {
      request( null, { uri: '/tags/' + conf.tag1._id }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I request one tag that doesn\'t exist': {
    topic: function () {
      request( userpass, {
        uri: '/tags/abcd1234'
      }, this.callback );
    },
    'THEN I should get a 404': function ( err, res, body ) {
      assert.equal( res.statusCode, 404 );
    }
  },

  'WHEN I request one tag': {
    topic: function () {
      request( userpass, {
        uri: '/tags/' + conf.tag1._id
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct tag': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data._id, conf.tag1._id );
      assert.equal( body.data.url, conf.tag1.url );
      assert.equal( body.data.blog, conf.blog1._id );
      assert.notEqual( String( new Date( body.data.created ) ), 'Invalid Date' );
      assert.notEqual( String( new Date( body.data.updated ) ), 'Invalid Date' );
    }
  },

  'WHEN I create a tag without authenticating': {
    topic: function () {
      request( null, {
        uri: '/tags',
        method: 'POST',
        body: JSON.stringify({
          url: 'bean',
          title: 'Beanz',
          blog: conf.blog1._id
        })
      }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I create a tag with no url': {
    topic: function () {
      request( userpass, {
        uri: '/tags',
        method: 'POST',
        body: JSON.stringify({
          title: 'Beanz',
          blog: conf.blog1._id
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I create a tag with no title': {
    topic: function () {
      request( userpass, {
        uri: '/tags',
        method: 'POST',
        body: JSON.stringify({
          url: 'george',
          blog: conf.blog1._id
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I create a tag with an invalid blog': {
    topic: function () {
      request( userpass, {
        uri: '/tags',
        method: 'POST',
        body: JSON.stringify({
          url: 'steveybean',
          title: 'Beanz',
          blog: 'foobarz'
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I create a tag with a created date': {
    topic: function () {
      var now = Date.now();
      request( userpass, {
        uri: '/tags',
        method: 'POST',
        body: JSON.stringify({
          url: 'steveybean',
          title: 'Beanz',
          blog: conf.blog1._id,
          created: now
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I create a tag with an updated date': {
    topic: function () {
      var now = Date.now();
      request( userpass, {
        uri: '/tags',
        method: 'POST',
        body: JSON.stringify({
          url: 'steveybean',
          title: 'Beanz',
          blog: conf.blog1._id,
          updated: now
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I create a tag': {
    topic: function () {
      request( userpass, {
        uri: '/tags',
        method: 'POST',
        body: JSON.stringify({
          url: 'steveybean',
          title: 'Beanz',
          blog: conf.blog1._id
        })
      }, this.callback );
    },
    'THEN I should get a 201': function ( err, res, body ) {
      assert.equal( res.statusCode, 201 );
    },
    'THEN I should get the tag back': function ( err, res, body ) {
      body = JSON.parse( body );

      created_id = body.data._id;

      assert.isNotNull( body.data._id );
      assert.equal( body.data.url, 'steveybean' );
      assert.equal( body.data.title, 'Beanz' );
      assert.equal( body.data.blog, conf.blog1._id );
      assert.notEqual( String( new Date( body.data.created ) ), 'Invalid Date' );
      assert.notEqual( String( new Date( body.data.updated ) ), 'Invalid Date' );
    }
  }

}).addBatch({

  'WHEN I verify that the tag was created': {
    topic: function () {
      Tag.findOne({ url: 'steveybean' })
        .populate( 'blog' )
        .run( this.callback );
    },
    'THEN I should get the correct data': function ( err, tag ) {
      assert.equal( tag._id, created_id );
      assert.equal( tag.url, 'steveybean' );
      assert.equal( tag.title, 'Beanz' );
      assert.equal( String( tag.blog._id ), String( conf.blog1._id ) );
      assert.isTrue( tag.created instanceof Date );
      assert.isTrue( tag.updated instanceof Date );
    }
  }

}).addBatch({

  'WHEN I update a tag without authenticating': {
    topic: function () {
      request( null, {
        uri: '/tags/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          url: 'george'
        })
      }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I update a tag that doesn\'t exist': {
    topic: function () {
      request( userpass, {
        uri: '/tags/abcd1234',
        method: 'PUT',
        body: JSON.stringify({
          url: 'george'
        })
      }, this.callback );
    },
    'THEN I should get a 404': function ( err, res, body ) {
      assert.equal( res.statusCode, 404 );
    }
  },

  'WHEN I update a tag with an invalid url': {
    topic: function () {
      request( userpass, {
        uri: '/tags/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          url: ''
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I update a tag with an invalid title': {
    topic: function () {
      request( userpass, {
        uri: '/tags/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          title: ''
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I update a tag with an invalid blog': {
    topic: function () {
      request( userpass, {
        uri: '/tags/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          blog: 'foobarbaz'
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I update a tag with a created date': {
    topic: function () {
      var now = Date.now();
      request( userpass, {
        uri: '/tags/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          created: now
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I update a tag with an updated date': {
    topic: function () {
      var now = Date.now();
      request( userpass, {
        uri: '/tags/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          updated: now
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I update a tag': {
    topic: function () {
      request( userpass, {
        uri: '/tags/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          url: 'george'
        })
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data back': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data._id, created_id );
      assert.equal( body.data.url, 'george' );
      assert.equal( body.data.blog, conf.blog1._id );
    }
  }

}).addBatch({

  // TODO verify that updated date was really updated

  'WHEN I verify that the tag was updated': {
    topic: function () {
      Tag.findById( created_id )
        .run( this.callback );
    },
    'THEN I should get the correct data': function ( err, tag ) {
      assert.equal( tag._id, created_id );
      assert.equal( tag.url, 'george' );
      assert.equal( tag.title, 'Beanz' );
      assert.equal( String( tag.blog ), String( conf.blog1._id ) );
    }
  }

}).addBatch({

  'WHEN I delete a tag without authenticating': {
    topic: function () {
      request( null, {
        uri: '/tags/' + created_id,
        method: 'DELETE',
      }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I delete a tag that doesn\'t exist': {
    topic: function () {
      request( userpass, {
        uri: '/tags/abcd1234',
        method: 'DELETE'
      }, this.callback );
    },
    'THEN I should get a 404': function ( err, res, body ) {
      assert.equal( res.statusCode, 404 );
    }
  },

  'WHEN I delete a tag': {
    topic: function () {
      request( userpass, {
        uri: '/tags/' + created_id,
        method: 'DELETE'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    }
  }

}).addBatch({

  'WHEN I verify that the tag was deleted': {
    topic: function () {
      Tag.findById( created_id, this.callback );
    },
    'THEN it should be gone': function ( err, doc ) {
      assert.isNull( doc );
    }
  },

  teardown: conf.teardown

}).export( module );

});

