var vows = require( 'vows' ),
    assert = require( 'assert' ),
    lib = require( './lib' ),
    request = lib.request,
    async = require( 'async' ),
    querystring = require( 'querystring' ),

    Author = require( '../models/Author' ),
    User = require( '../models/User' );

function setup ( cb ) {
  lib.setup( function ( conf ) {

  var author1 = new Author({
    name: 'testauthor1',
    url: 'test-author-1'
  });

  author1.save( function () {
    var user1 = new User({
      name: 'testuser1',
      password: '1234',
      admin: true,
      author: author1
    });

    var user2 = new User({
      name: 'testuser2',
      password: '1234',
      author: author1
    });

    user1.save( function () {
      user2.save( function () {
        conf.author1 = author1;
        conf.user1 = user1;
        conf.user2 = user2;

        cb( conf );
      });
    });
  });

  });
}

setup( function ( conf ) {

var userpass = [ conf.user, conf.password ],
  created_id;

function verify( data ) {
  var i = 0, found = 0;
  for ( ; i < data.length; i ++ ) {
    if ( data[ i ].name.substring( 0, 8 ) === 'testuser' ) {
      found++;
    }
    if ( data[ i ].password ) {
      assert.fail();
    }
  }
  return found;
}

vows.describe( 'User Tests' ).addBatch({

  'WHEN I list users without authenticating': {
    topic: function () {
      request( null, { uri: '/users' }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I list users': {
    topic: function () {
      request( userpass, { uri: '/users' }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      var found = verify( body.data );

      assert.equal( found, 3 );
      assert.equal( body.data.length, 3 );
    }
  },

  'WHEN I list users sorted': {
    topic: function () {
      request( userpass, {
        uri: '/users?sort=name&dir=desc'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      var found = verify( body.data );

      assert.equal( found, 3 );
      assert.equal( body.data.length, 3 );
    },
    'THEN the order should be correct': function ( err, res, body ) {
      body = JSON.parse( body );

      for ( var i = 0; i < body.data.length; i++ ) {
        if ( i === 0 ) {
          continue;
        } else if ( body.data[ i - 1 ].name < body.data[ i ].name ) {
          assert.fail();
        }
      }
    }
  },

  'WHEN I list users sorted by created date': {
    topic: function () {
      request( userpass, {
        uri: '/users?sort=created&dir=asc'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN the data should be correct': function ( err, res, body ) {
      body = JSON.parse( body );

      var found = verify( body.data );

      assert.equal( found, 3 );
      assert.equal( body.data.length, 3 );
    },
    'THEN the data should be sorted correctly': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( String( body.data[ 1 ]._id ), String( conf.user1._id ) );
      assert.equal( String( body.data[ 2 ]._id ), String( conf.user2._id ) );
    }
  },

  'WHEN I query sorted and starting at item 1': {
    topic: function () {
      request( userpass, {
        uri: '/users?sort=name&dir=asc&start=1'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data.length, 2 );
      assert.equal( body.data[ 0 ].name, 'testuser1' );
      assert.equal( body.data[ 1 ].name, 'testuser2' );
    }
  },

  'WHEN I query with a limit': {
    topic: function () {
      request( userpass, {
        uri: '/users?limit=1'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get only one item': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data.length, 1 );
    }
  },

  'WHEN I query with eq': {
    topic: function () {
      var query = querystring.stringify({
        filter: JSON.stringify([
          [ '_id', 'eq', conf.user1._id ]
        ])
      });
      request( userpass, {
        uri: '/users?' + query
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data[ 0 ]._id, conf.user1._id );
    }
  },

  'WHEN I query with gt': {
    topic: function () {
      var query = querystring.stringify({
        filter: JSON.stringify([
          [ 'created', 'gt', conf.user1.created ]
        ])
      });
      request( userpass, {
        uri: '/users?' + query
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data[ 0 ]._id, conf.user2._id );
    }
  },

  'WHEN I query with gte': {
    topic: function () {
      var query = querystring.stringify({
        filter: JSON.stringify([
          [ 'created', 'gte', conf.user1.created ]
        ]),
        sort: 'created',
        dir: 'asc'
      });
      request( userpass, {
        uri: '/users?' + query
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data[ 0 ]._id, conf.user1._id );
      assert.equal( body.data[ 1 ]._id, conf.user2._id );
    }
  },

  'WHEN I query with in': {
    topic: function () {
      var query = querystring.stringify({
        filter: JSON.stringify([
          [ 'name', 'in', [ 'beanz', 'testuser1' ] ]
        ])
      });
      request( userpass, {
        uri: '/users?' + query
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data[ 0 ]._id, conf.user1._id );
    }
  },

  'WHEN I query with lt': {
    topic: function () {
      var query = querystring.stringify({
        filter: JSON.stringify([
          [ 'created', 'lt', conf.user2.created ]
        ])
      });
      request( userpass, {
        uri: '/users?' + query
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( String( body.data[ 1 ]._id ), String( conf.user1._id ) );
    }
  },

  'WHEN I query with lte': {
    topic: function () {
      var query = querystring.stringify({
        filter: JSON.stringify([
          [ 'created', 'lte', conf.user2.created ]
        ]),
        sort: 'created',
        dir: 'asc'
      });
      request( userpass, {
        uri: '/users?' + query
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( String( body.data[ 1 ]._id ), String( conf.user1._id ) );
      assert.equal( String( body.data[ 2 ]._id ), String( conf.user2._id ) );
    }
  },

  'WHEN I query with match': {
    topic: function () {
      var query = querystring.stringify({
        filter: JSON.stringify([
          [ 'name', 'match', 'user1' ]
        ])
      });
      request( userpass, {
        uri: '/users?' + query
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data[ 0 ]._id, conf.user1._id );
    }
  },

  'WHEN I request one user without authenticating': {
    topic: function () {
      request( null, { uri: '/users/' + conf.user1._id }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I request one user that doesn\'t exist': {
    topic: function () {
      request( userpass, {
        uri: '/users/abcd1234'
      }, this.callback );
    },
    'THEN I should get a 404': function ( err, res, body ) {
      assert.equal( res.statusCode, 404 );
    }
  },

  'WHEN I request one user (admin)': {
    topic: function () {
      request( userpass, {
        uri: '/users/' + conf.user1._id
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct user': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data._id, conf.user1._id );
      assert.equal( body.data.name, conf.user1.name );
      assert.equal( body.data.admin, true );
      assert.equal( body.data.author, conf.author1._id );
      assert.isUndefined( body.data.password );
      assert.notEqual( String( new Date( body.data.created ) ), 'Invalid Date' );
      assert.notEqual( String( new Date( body.data.updated ) ), 'Invalid Date' );
    }
  },

  'WHEN I request one user (non-admin)': {
    topic: function () {
      request( userpass, {
        uri: '/users/' + conf.user2._id
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct user': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data._id, conf.user2._id );
      assert.equal( body.data.name, conf.user2.name );
      assert.equal( body.data.admin, false );
      assert.equal( body.data.author, conf.author1._id );
      assert.isUndefined( body.data.password );
      assert.notEqual( String( new Date( body.data.created ) ), 'Invalid Date' );
      assert.notEqual( String( new Date( body.data.updated ) ), 'Invalid Date' );
    }
  },

  'WHEN I create a user without authenticating': {
    topic: function () {
      request( null, {
        uri: '/users',
        method: 'POST',
        body: JSON.stringify({
          name: 'bean',
          password: 'password',
          author: conf.author1._id
        })
      }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I create a user with a conflicting username': {
    topic: function () {
      request( userpass, {
        uri: '/users',
        method: 'POST',
        body: JSON.stringify({
          name: 'testuser1',
          password: 'password',
          author: conf.author1._id
        })
      }, this.callback );
    },
    'THEN I should get a 409': function ( err, res, body ) {
      assert.equal( res.statusCode, 409 );
    }
  },

  'WHEN I create a user with no name': {
    topic: function () {
      request( userpass, {
        uri: '/users',
        method: 'POST',
        body: JSON.stringify({
          password: 'password',
          author: conf.author1._id
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I create a user with no password': {
    topic: function () {
      request( userpass, {
        uri: '/users',
        method: 'POST',
        body: JSON.stringify({
          name: 'steveybean',
          author: conf.author1._id
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I create a user with an invalid author': {
    topic: function () {
      request( userpass, {
        uri: '/users',
        method: 'POST',
        body: JSON.stringify({
          name: 'steveybean',
          password: 'password',
          author: 'foobarbaz'
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I create a user with a created date': {
    topic: function () {
      var now = Date.now();
      request( userpass, {
        uri: '/users',
        method: 'POST',
        body: JSON.stringify({
          name: 'steveybean',
          password: 'password',
          author: conf.author1._id,
          created: now
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I create a user with an updated date': {
    topic: function () {
      var now = Date.now();
      request( userpass, {
        uri: '/users',
        method: 'POST',
        body: JSON.stringify({
          name: 'steveybean',
          password: 'password',
          author: conf.author1._id,
          updated: now
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I create a user': {
    topic: function () {
      request( userpass, {
        uri: '/users',
        method: 'POST',
        body: JSON.stringify({
          name: 'steveybean',
          password: userpass[ 1 ],
          author: conf.author1._id
        })
      }, this.callback );
    },
    'THEN I should get a 201': function ( err, res, body ) {
      assert.equal( res.statusCode, 201 );
    },
    'THEN I should get the user back': function ( err, res, body ) {
      body = JSON.parse( body );

      created_id = body.data._id;

      assert.isNotNull( body.data._id );
      assert.equal( body.data.name, 'steveybean' );
      assert.equal( body.data.admin, false );
      assert.equal( body.data.author, conf.author1._id );
      assert.isUndefined( body.data.password );
      assert.notEqual( String( new Date( body.data.created ) ), 'Invalid Date' );
      assert.notEqual( String( new Date( body.data.updated ) ), 'Invalid Date' );
    }
  }

}).addBatch({

  'WHEN I verify that the user was created': {
    topic: function () {
      User.findOne({ name: 'steveybean' })
        .populate( 'author' )
        .run( this.callback );
    },
    'THEN I should get the correct data': function ( err, user ) {
      assert.equal( user._id, created_id );
      assert.equal( user.name, 'steveybean' );
      assert.equal( user.admin, false );
      assert.equal( user.password, userpass[ 0 ].password );
      assert.equal( String( user.author._id ), String( conf.author1._id ) );
      assert.isTrue( user.created instanceof Date );
      assert.isTrue( user.updated instanceof Date );
    }
  }

}).addBatch({

  'WHEN I update a user without authenticating': {
    topic: function () {
      request( null, {
        uri: '/users/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          name: 'george'
        })
      }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I update a user that doesn\'t exist': {
    topic: function () {
      request( userpass, {
        uri: '/users/1234',
        method: 'PUT',
        body: JSON.stringify({
          name: 'george'
        })
      }, this.callback );
    },
    'THEN I should get a 404': function ( err, res, body ) {
      assert.equal( res.statusCode, 404 );
    }
  },

  'WHEN I update a user with an invalid name': {
    topic: function () {
      request( userpass, {
        uri: '/users/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          name: ''
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I update a user with an invalid password': {
    topic: function () {
      request( userpass, {
        uri: '/users/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          password: ''
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  // TODO test updating with invalid author

  'WHEN I update a user with a created date': {
    topic: function () {
      var now = Date.now();
      request( userpass, {
        uri: '/users/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          created: now
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },

  'WHEN I update a user with an updated date': {
    topic: function () {
      var now = Date.now();
      request( userpass, {
        uri: '/users/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          updated: now
        })
      }, this.callback );
    },
    'THEN I should get a 403': function ( err, res, body ) {
      assert.equal( res.statusCode, 403 );
    }
  },


  'WHEN I update a user': {
    topic: function () {
      request( userpass, {
        uri: '/users/' + created_id,
        method: 'PUT',
        body: JSON.stringify({
          name: 'george',
          password: 'george'
        })
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    },
    'THEN I should get the correct data back': function ( err, res, body ) {
      body = JSON.parse( body );

      assert.equal( body.data._id, created_id );
      assert.equal( body.data.name, 'george' );
      assert.isUndefined( body.data.password );
    }
  }

}).addBatch({

  // TODO verify that updated date was really updated

  'WHEN I verify that the user was updated': {
    topic: function () {
      User.findById( created_id )
        .run( this.callback );
    },
    'THEN I should get the correct data': function ( err, user ) {
      assert.equal( user._id, created_id );
      assert.equal( user.name, 'george' );
    }
  }

}).addBatch({

  'WHEN I delete a user without authenticating': {
    topic: function () {
      request( null, {
        uri: '/users/' + created_id,
        method: 'DELETE',
      }, this.callback );
    },
    'THEN I should get a 401': function ( err, res, body ) {
      assert.equal( res.statusCode, 401 );
    }
  },

  'WHEN I delete a user that doesn\'t exist': {
    topic: function () {
      request( userpass, {
        uri: '/users/abcd1234',
        method: 'DELETE'
      }, this.callback );
    },
    'THEN I should get a 404': function ( err, res, body ) {
      assert.equal( res.statusCode, 404 );
    }
  },

  'WHEN I delete a user': {
    topic: function () {
      request( userpass, {
        uri: '/users/' + created_id,
        method: 'DELETE'
      }, this.callback );
    },
    'THEN I should get a 200': function ( err, res, body ) {
      assert.equal( res.statusCode, 200 );
    }
  }

}).addBatch({

  'WHEN I verify that the user was deleted': {
    topic: function () {
      User.findById( created_id, this.callback );
    },
    'THEN it should be gone': function ( err, doc ) {
      assert.isNull( doc );
    }
  },

  teardown: conf.teardown

}).export( module );

});

